/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "AlignmentLogo.h"

#include <util_ov_msaedit/MSAEditor.h>

#include <QtGui/QPainter>
#include <QHBoxLayout>
#include "datatype/MAlignment.h"
#include "core_api/DNAAlphabet.h"

namespace GB2 {

/************************************************************************/
/* LogoRenderArea                                                       */
/************************************************************************/
AlignmentLogoRenderArea::AlignmentLogoRenderArea(const AlignmentLogoSettings& _s, QWidget* p)
: QWidget(p), settings(_s) {
    QHBoxLayout* layout = new QHBoxLayout();
    layout->addWidget(this);
    p->setLayout(layout);

    bases<<'A'<<'G'<<'C'<<'T'<<'U';
    /*aminoacids<<'A'<<'C'<<'D'<<'E'<<'F'<<'G'<<'H'
        <<'I'<<'K'<<'L'<<'M'<<'N'<<'P'<<'Q'<<'R'
        <<'S'<<'T'<<'V'<<'W'<<'Y';*/

    acceptableChars = new QVector<char>();
    switch (settings.sequenceType)
    {
        case NA:
            acceptableChars = &bases;
            s = 4.0;
            break;
        default:
            QByteArray chars = settings.ma.getAlphabet()->getAlphabetChars();
            foreach(char ch, chars) {
                if(ch!=MAlignment_GapChar)
                    acceptableChars->append(ch);
            }
            s = 20.0;
            //acceptableChars = &aminoacids;
            break;
    }

    evaluateHeights();
    sortCharsByHeight();
}

void AlignmentLogoRenderArea::replaceSettings(const AlignmentLogoSettings& _s) {
    settings = _s;

    acceptableChars = new QVector<char>();
    switch (settings.sequenceType)
    {
        case NA:
            acceptableChars = &bases;
            s = 4.0;
            break;
        default:
            QByteArray chars = settings.ma.getAlphabet()->getAlphabetChars();
            foreach(char ch, chars) {
                if(ch!=MAlignment_GapChar)
                    acceptableChars->append(ch);
            }
            s = 20.0;
            //acceptableChars = &aminoacids;
            break;
    }
    columns.clear();
    for (int i = 0; i < 256; i++) {
        frequencies[i].clear();
        heights[i].clear();
    }
    evaluateHeights();
    sortCharsByHeight();
}

#define SPACER 1
void AlignmentLogoRenderArea::paintEvent(QPaintEvent* e) {
    QPainter p(this);
    p.fillRect(0,0,width(),height(),Qt::white);
    QFont charFont("Helvetica");
    int bitSize = height() / 2;
    charFont.setPixelSize(bitSize);
    charFont.setBold(true);
    int maxHeight = height();
    QFontMetrics fm(charFont);
    int columnWidth = 0;
    for (int i = 0, n = acceptableChars->size(); i < n; i++) {
        columnWidth = qMax(columnWidth, fm.width(acceptableChars->at(i)));
    }

    if (columnWidth * settings.len > width()) {
        bitSize = bitSize * width() / (columnWidth * settings.len);
        charFont.setPixelSize(bitSize);
        int maxHeight = 2 * bitSize;
        columnWidth = 0;
        fm = QFontMetrics(charFont);
        for (int i = 0, n = acceptableChars->size(); i < n; i++) {
            columnWidth = qMax(columnWidth, fm.width(acceptableChars->at(i)));
        }
    }

    int yLevel=maxHeight;
    int colNum=0;
    for(int pos=0; pos < settings.len; pos++) {
        assert(pos < columns.size());
        const QVector<char>& charsAt = columns.at(pos);
        foreach(char ch, charsAt) {
            QPointF baseline(colNum * columnWidth, yLevel);
            int charHeight = heights[ch][pos] * bitSize;
            /*if(charHeight<2) {
                continue;
            }*/
            QColor charColor = Qt::black;
            if(settings.colorScheme.contains(ch)) {
                charColor = settings.colorScheme.value(ch);
            }
            AlignmentLogoItem* logoItem = new AlignmentLogoItem(ch, baseline, charHeight, charFont, charColor);
            logoItem->paint(&p, NULL, this);
            yLevel-=charHeight+SPACER;
        }
        yLevel=maxHeight;
        colNum++;
    }

    QWidget::paintEvent(e);
}

void AlignmentLogoRenderArea::evaluateHeights() {
    const MAlignment& ma = settings.ma;
    int numRows = ma.getNumRows();
    error = (s - 1)/(2*log(2.0)*numRows);

    foreach (char ch, *acceptableChars) {
        QVector<qreal> freqs(settings.len);
        QVector<qreal> hts(settings.len);
        frequencies[ch] = freqs;
        heights[ch] = hts;
    }
    columns.resize(settings.len);

    for (int pos = settings.startPos; pos < settings.len + settings.startPos; pos++) {
        for (int idx = 0; idx < numRows; idx++) {
            const MAlignmentRow& row = ma.getRow(idx);
            assert(pos < row.getCoreLength());
            char ch = row.chatAt(pos);
            if(acceptableChars->contains(ch)) {
                int arrIdx = pos - settings.startPos;
                assert(arrIdx >= 0);
                assert(arrIdx < frequencies[ch].size());
                frequencies[ch][arrIdx]+=1.0;
                if (!columns[arrIdx].contains(ch)) {
                    columns[arrIdx].append(ch);
                }
            }
        }
    }

    int rows = settings.ma.getNumRows();
    for(int pos=0; pos<settings.len; pos++) {
        qreal h = getH(pos);
        foreach(char c, columns[pos]) {
            qreal freq = frequencies[c][pos] / rows;
            heights[c][pos] = freq * ( log(s) / log(2.0) - ( h + error ) );
        }
    }
}

qreal AlignmentLogoRenderArea::getH(int pos) {
    qreal h = 0.0;
    int rows = settings.ma.getNumRows();
    foreach(char ch, columns.at(pos)) {
        qreal freq = frequencies[ch][pos] / rows;
        h += -freq * log(freq) / log(2.0);
    }
    return h;
}

void AlignmentLogoRenderArea::sortCharsByHeight() {
    for(int pos=0; pos<columns.size(); pos++) {
        QVector<char>& chars = columns[pos];
        char temp;
        int count = chars.size();
        for(int j = 0; j < chars.size() - 1; j++) {
            for(int i = 0; i < count - 1; i++) {
                temp = chars[i];
                qreal tempFreq = frequencies[temp][pos];
                qreal nextFreq = frequencies[chars[i+1]][pos];
                if (tempFreq>nextFreq) {
                    chars[i] = chars[i+1];
                    chars[i+1] = temp;
                }
                else {
                    temp = chars[i+1];
                }
            }
            --count;
        }
    }
}

/************************************************************************/
/* Logo item                                                            */
/************************************************************************/
AlignmentLogoItem::AlignmentLogoItem(char _ch, QPointF _baseline, int _charHeight, QFont _font, QColor _color)
: ch(_ch), baseline(_baseline), charHeight(_charHeight), font(_font), color(_color) {}

QRectF AlignmentLogoItem::boundingRect() const {
    return path.boundingRect();
}

void AlignmentLogoItem::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget ) {
    Q_UNUSED(option);
    Q_UNUSED(widget);

    painter->save();
    QString chStr(ch);
    path.addText(baseline, font, chStr);

    //adjust item's height
    QRectF bound = path.boundingRect();

    int charWidth = 30;

    qreal sx = charWidth / bound.width();
    qreal sy = charHeight / bound.height();
    
    painter->scale(sx, sy);

    //map baseline position to scaled coordinates
    qreal offsetx = baseline.x() * (1/sx - 1);
    qreal offsety = baseline.y() * (1/sy - 1);
    painter->translate(offsetx, offsety);

    painter->fillPath(path, color);
    painter->restore();
}

}//namespace
