/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include <QtCore/QCryptographicHash>
#include <QtNetwork/QNetworkRequest>

#include <core_api/AppContext.h>
#include <core_api/NetworkConfiguration.h>
#include <core_api/AppSettings.h>
#include <core_api/Log.h>
#include <core_api/GUrl.h>

#include <distributed_computing/RemoteMachineMonitor.h>
#include <workflow_support/CoreLibConstants.h>

#include "CloudSupport.h"
#include "CloudRemoteMachine.h"
#include "CloudTransportProtocol.h"
#include "UctpRequestBuilders.h"

namespace GB2 {

#define ULOG_CAT_CLOUD_REMOTE_MACHINE "cloud-remote-machine"
static LogCategory log(ULOG_CAT_CLOUD_REMOTE_MACHINE);


// CloudRemoteMachine

CloudRemoteMachine::CloudRemoteMachine(CloudRemoteMachineSettings* s):
    settings(s), eventLoop( NULL ), reply( NULL ), protocolHandler(new Uctp()), session(NULL), errorOccured(false)
{

    cloudUrl = settings->getUrl();
   
    NetworkConfiguration* nc = AppContext::getAppSettings()->getNetworkConfiguration();
    proxy = nc->getProxyByUrl(cloudUrl);

    QByteArray sid = settings->getSessionId().toAscii();
    if (!sid.isEmpty() ) {
        session.reset(new UctpSession(sid));    
    }
    log.trace("Started cloud remote machine instance");

}

CloudRemoteMachine::~CloudRemoteMachine()
{

}


RemoteTaskError CloudRemoteMachine::runRemoteTask( const QString & taskFactoryId, const QVariant & settings, qint64 * taskId )
{
 
    // Only workflow tasks! 
    if (taskFactoryId != Workflow::CoreLibConstants::WORKFLOW_ON_CLOUD_TASK_ID) {
        return RemoteTaskError(false, QString("Unsupported task type"));
    }    

    RemoteTaskError result = initSession();
    if (false == result.getOk()) {
        return RemoteTaskError(false, QString("Failed to init session (%1)").arg(result.getMsg()));
    }
    log.trace("Init session successful");

    QVariantMap settingsMap = settings.toMap();
    const QByteArray schema = settingsMap.value(Workflow::CoreLibConstants::WORKFLOW_SCHEMA_ATTR).toByteArray();
    const QStringList inUrls =  settingsMap.value(Workflow::CoreLibConstants::DATA_IN_ATTR).toStringList();
    const QStringList outUrls = settingsMap.value(Workflow::CoreLibConstants::DATA_OUT_ATTR).toStringList();
    
    try {
        RunRemoteTaskRequest request(UctpCommands::RUN_TASK, session.get(), schema, inUrls, outUrls);
        QByteArray data = protocolHandler->formRequest(request);
        sendRequest(request.getCommand(), data);
        if (!errorOccured) {
            // id looks like task########, so we omit first 4 chars
            QString taskIdStr = replyData.value(UctpElements::TASK_ID).toString().mid(4);
            bool ok = false;
            *taskId = taskIdStr.toLongLong(&ok,16);
            if (!ok) {
                errorOccured = true;
                errorMessage = QString("Incorrect task id: %1").arg(*taskId);
            }
        } 
    } catch (UctpError& e) {
        return RemoteTaskError( false, e.getErrorMessage());
    }

    return RemoteTaskError(!errorOccured, errorMessage);
}


RemoteTaskError CloudRemoteMachine::cancelRemoteTask( qint64 taskId )
{
    UpdateRemoteTaskRequest request(UctpCommands::CANCEL_TASK, session.get(), taskId);
    
    QByteArray data = protocolHandler->formRequest(request);
    sendRequest(request.getCommand(), data);
    
    return RemoteTaskError(!errorOccured, errorMessage);
}

RemoteTaskError CloudRemoteMachine::deleteRemoteTask( qint64 taskId )
{

    return RemoteTaskError(false, tr("Delete remote task is not implemented"));
}

RemoteTaskError CloudRemoteMachine::getRemoteTaskCancelFlag( qint64 taskId, bool * cancelFlag )
{
    return RemoteTaskError(false,tr("Get remote task flag is not implemented."));
}


RemoteTaskError CloudRemoteMachine::getRemoteTaskState( qint64 taskId, Task::State * state )
{
    assert(state != NULL);
     
    QStringList properties;
    properties.append(UctpElements::TASK_STATE);
    properties.append(UctpElements::TASK_ERROR);
    GetRemoteTaskStatusRequest request(UctpCommands::GET_TASK_PROPERTY, session.get(), taskId, properties);

    QByteArray data = protocolHandler->formRequest(request);
    sendRequest(request.getCommand(), data);

    if (!errorOccured) {
        QByteArray stateText = replyData.value(UctpElements::TASK_STATE).toByteArray();
        if (stateText == UctpValues::UCTP_STATE_FINISHED) {
            *state = Task::State_Finished;
            QByteArray errorText = replyData.value(UctpElements::TASK_ERROR).toByteArray();
            if (!errorText.isEmpty()) {
                errorOccured = true;
                errorMessage = errorText;
            }
        } else if (stateText.isEmpty()) {
            errorOccured = true;
            errorMessage = tr("Empty task state property");
        }
    }
    
    return RemoteTaskError(!errorOccured, errorMessage);
}

RemoteTaskError CloudRemoteMachine::getRemoteTaskProgress( qint64 taskId, int * progress )
{
    assert(progress != NULL);
    
    QStringList properties;
    properties.append(UctpElements::TASK_PROGRESS);
    GetRemoteTaskStatusRequest request(UctpCommands::GET_TASK_PROPERTY, session.get(), taskId, properties);

    QByteArray data = protocolHandler->formRequest(request);
    sendRequest(request.getCommand(), data);

    if (!errorOccured) {
        QString buf = replyData.value(UctpElements::TASK_PROGRESS).toString();
        *progress = buf.toInt();
    }

    return RemoteTaskError(!errorOccured, errorMessage);
}

RemoteTaskError CloudRemoteMachine::getRemoteTaskResult( qint64 taskId, QVariant * resultUrls )
{
    assert(resultUrls != NULL);

    try {

        QStringList urls = resultUrls->toStringList();
        if (urls.isEmpty()) {
            return RemoteTaskError(!errorOccured, errorMessage);
        }
        GetRemoteTaskStatusRequest request(UctpCommands::GET_TASK_RESULT, session.get(), taskId);

        QByteArray data = protocolHandler->formRequest(request);
        sendRequest(request.getCommand(), data);

        if (!errorOccured) {
            foreach( const GUrl& url, urls ) {
                QString dataName = url.fileName();
                if (!replyData.contains(dataName)) 
                {
                    errorOccured = true;
                    errorMessage = tr("Result {%1} is not found.").arg(dataName);
                }
                const QByteArray data = replyData.value(dataName).toByteArray();
                saveResult( QByteArray::fromBase64(data), url.getURLString());
            }
        }
    } catch (UctpError& e) {
        return RemoteTaskError(false, e.getErrorMessage());
    }

    return RemoteTaskError(!errorOccured, errorMessage);
}

RemoteTaskError CloudRemoteMachine::getRemoteTaskError( qint64 taskId, QString * errMsg )
{
    assert(errMsg != NULL);
    
    QStringList properties;
    properties.append(UctpElements::TASK_ERROR);
    GetRemoteTaskStatusRequest request(UctpCommands::GET_TASK_PROPERTY, session.get(), taskId,properties);

    QByteArray data = protocolHandler->formRequest(request);
    sendRequest(request.getCommand(), data);

    if (!errorOccured) {
        *errMsg = replyData.value(UctpElements::TASK_ERROR).toString();
    }

    return RemoteTaskError(!errorOccured, errorMessage);
}

RemoteTaskError CloudRemoteMachine::getUuid( QUuid * uuid )
{
    assert(uuid != NULL);

    RemoteTaskError result = initSession();
    if (false == result.getOk()) {
        return RemoteTaskError(false, QString("Failed to init session (%1)").arg(result.getMsg()));
    }
    
    session->buildQUuid(uuid);

    return RemoteTaskError(!errorOccured, errorMessage);
}

#define WORKFLOW_REMOTE_RUN_ATTR "Workflow schema simple run task"

RemoteTaskError CloudRemoteMachine::getServicesList( QStringList * list ) {
    
    assert(list != NULL);
    
     RemoteTaskError result = initSession();
     if (false == result.getOk()) {
         return RemoteTaskError(false, QString("Failed to init session (%1)").arg(result.getMsg()));
     }
     
//     GetServiceListRequest request(UctpCommands::GET_SERVICE_LIST);
// 
//     QByteArray data = protocolHandler->formRequest(request);
//     sendRequest(request.getCommand(), data);
    
    if (!errorOccured) {
        // old legacy
        list->clear();
        list->append(WORKFLOW_REMOTE_RUN_ATTR);
    }
    
    return RemoteTaskError(!errorOccured, errorMessage);
}

RemoteTaskError CloudRemoteMachine::getHostName( QString * hostname ) {

    assert (hostname != NULL);

    RemoteTaskError result = initSession();
    if (false == result.getOk()) {
        return RemoteTaskError(false, QString("Failed to init session (%1)").arg(result.getMsg()));
    }
    
    PingRequest request(UctpCommands::PING);

    QByteArray data = protocolHandler->formRequest(request);
    sendRequest(request.getCommand(), data);
    
    if (!errorOccured) {
        log.trace("Ping remote machine successful");
        *hostname = "Amazon EC2 ";
    }

    return RemoteTaskError(!errorOccured, errorMessage);
}


void CloudRemoteMachine::sl_onReply( )
{
    assert(reply != NULL);
    assert(eventLoop != NULL);

    if (reply->error() == QNetworkReply::NoError) {
        bool parseOk = protocolHandler->parseReply(reply, activeCommand, replyData);
        if (!parseOk) {
            errorOccured = true;
            errorMessage = QString(tr("Failed to parse server response. %1")).arg(protocolHandler->getErrorText());
        }

    } else {
        errorOccured = true;
        errorMessage = reply->errorString();
    }
    
    eventLoop->exit();
}


void CloudRemoteMachine::sendRequest( const QByteArray& command, const QByteArray& data )
{
    errorMessage.clear();
    errorOccured = false;
    activeCommand = command;
  
    if (data.isEmpty()) {
        errorOccured = true;
        errorMessage = QString(tr("Send request failed: empty message data"));
        return;
    }

    eventLoop = new QEventLoop();
    QNetworkAccessManager* networkManager = new QNetworkAccessManager();

    QNetworkRequest request (cloudUrl);
    request.setHeader(QNetworkRequest::ContentTypeHeader, "application/xml");
    request.setHeader(QNetworkRequest::ContentLengthHeader, data.length());
    reply = networkManager->post(request, data );
    connect(reply, SIGNAL(finished()), SLOT(sl_onReply()));

    eventLoop->exec();

    delete eventLoop;
    delete networkManager;
}

static QByteArray getMD5Hash(const QByteArray& data) {
    QCryptographicHash hash(QCryptographicHash::Md5);
    hash.addData(data);
    return hash.result();
}

RemoteTaskError CloudRemoteMachine::initSession()
{
    if (session.get() == NULL) {
        QByteArray userName = settings->getUserName().toAscii();
        QByteArray pass = settings->getPasswd().toAscii();
        
        InitSessionRequest request(UctpCommands::AUTH, userName, pass);
        
        QByteArray data = protocolHandler->formRequest(request);
        sendRequest(request.getCommand(), data);
        if (!errorOccured) {
            QByteArray id = replyData.value(UctpElements::SESSION_ID).toByteArray();
            session.reset(new UctpSession(id));
            updateGlobalSettings();
        }
    }
    
    return RemoteTaskError(!errorOccured, errorMessage);
}

void CloudRemoteMachine::saveResult( const QByteArray& result, const QString& path )
{
    QFile resultFile(path);
    if ( !resultFile.open(QIODevice::WriteOnly) ) {
        throw UctpError("Can not open file to write results");
    }

    qint64 bytesWrittern = resultFile.write(result);
    if (bytesWrittern != result.size()) {
        throw UctpError("Error writing results");
    }
}

void CloudRemoteMachine::updateGlobalSettings()
{
    settings->setSessionId(session->getUid());
}



// CloudRemoteMachineSettings

CloudRemoteMachineSettings::CloudRemoteMachineSettings():
RemoteMachineSettings( AppContext::getProtocolInfoRegistry()->getProtocolInfo( CloudSupport::CLOUD_SUPPORT_PROTOCOL_ID), RemoteMachineType_CloudServer)

{
    log.trace("Cloud machine settings created");   
}

CloudRemoteMachineSettings::CloudRemoteMachineSettings(const QString &host, const QString& user, const QString& pass):
RemoteMachineSettings(AppContext::getProtocolInfoRegistry()->getProtocolInfo( CloudSupport::CLOUD_SUPPORT_PROTOCOL_ID ), RemoteMachineType_CloudServer ),
url(host), name(user), passwd(pass)
{
    log.trace("Cloud machine settings created");
}

QString CloudRemoteMachineSettings::toString() const
{
    
    return QString("%1;%2;%3").arg(url).arg(name).arg(passwd);
}


bool CloudRemoteMachineSettings::operator ==( const RemoteMachineSettings & m ) const {
    const RemoteMachineSettings * machine = &m;
    const CloudRemoteMachineSettings* cfg = dynamic_cast< const CloudRemoteMachineSettings* >( machine );
    if( NULL == cfg ) {
        return false;
    }
    return url == cfg->getUrl() && name == cfg->getUserName() && passwd == cfg->getPasswd();
}

QString CloudRemoteMachineSettings::serialize() const
{
    return toString();
}

bool CloudRemoteMachineSettings::deserialize( const QString & data )
{
    QStringList values = data.split(";", QString::SkipEmptyParts);
    if (values.count() != 3) {
        return false;
    }
    url = values.at(0);
    name = values.at(1);
    passwd = values.at(2);

    return true;
}

// CloudRemoteMachineFactory

CloudRemoteMachineFactory::CloudRemoteMachineFactory()
{
}

CloudRemoteMachineFactory::~CloudRemoteMachineFactory()
{
}

RemoteMachine * CloudRemoteMachineFactory::createInstance( const QString & serializedSettings ) const
{
    CloudRemoteMachineSettings settings;
    if(settings.deserialize(serializedSettings))
    {
        return createInstance(&settings);
    }
    return NULL;
}

RemoteMachine * CloudRemoteMachineFactory::createInstance( RemoteMachineSettings * settings ) const
{
    CloudRemoteMachineSettings *castedSettings = dynamic_cast<CloudRemoteMachineSettings *>(settings);
    if(NULL != castedSettings)
    {
        return new CloudRemoteMachine(castedSettings);
    }
    return NULL;
}

RemoteMachineSettings * CloudRemoteMachineFactory::createSettings( const QString & serializedSettings ) const {
    CloudRemoteMachineSettings * ret = new CloudRemoteMachineSettings();
    if( ret->deserialize( serializedSettings ) ) {
        return ret;
    }
    delete ret;
    return NULL;
}

} // namespace GB2
