/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include <core_api/Log.h>
#include "CloudTransportProtocol.h"

namespace GB2 {

const QByteArray UctpElements::REQUEST("uctp:Request");
const QByteArray UctpElements::RESPONSE("uctp:Response");
const QByteArray UctpElements::HEADER("uctp:Header");
const QByteArray UctpElements::CONTENTS("uctp:Contents");
const QByteArray UctpElements::SESSION_ID("session-id");
const QByteArray UctpElements::RESULT("result");
const QByteArray UctpElements::REASON("reason");
const QByteArray UctpElements::USER_NAME("username");
const QByteArray UctpElements::PASSWD("passwd");
const QByteArray UctpElements::PROPERTY("property");
const QByteArray UctpElements::TASK_SCHEMA("task");
const QByteArray UctpElements::TASK_ID("task-id");
const QByteArray UctpElements::TASK_STATE("state");
const QByteArray UctpElements::TASK_ERROR("error");
const QByteArray UctpElements::TASK_PROGRESS("progress");
const QByteArray UctpElements::DATA("data");


const QByteArray UctpAttributes::COMMAND_TYPE("type");
const QByteArray UctpAttributes::IS_GZIPPED("gzipped");
const QByteArray UctpAttributes::ENCODING("encoding");
const QByteArray UctpAttributes::NAME("name");
const QByteArray UctpAttributes::SIZE("size");

const UctpCommand UctpCommands::AUTH("AUTH");
const UctpCommand UctpCommands::RUN_TASK("RUN_TASK");
const UctpCommand UctpCommands::PING("PING");
const UctpCommand UctpCommands::GET_TASK_RESULT("GET_TASK_RESULT");
const UctpCommand UctpCommands::GET_TASK_PROPERTY("GET_TASK_PROPERTY");
const UctpCommand UctpCommands::CANCEL_TASK("CANCEL_TASK");

const QByteArray UctpValues::BASE64("BASE64");
const QByteArray UctpValues::UCTP_TRUE("true");
const QByteArray UctpValues::UCTP_FALSE("false");
const QByteArray UctpValues::UCTP_STATE_FINISHED("FINISHED");
const QByteArray UctpValues::UCTP_STATE_ERROR("ERROR");
const QByteArray UctpValues::UCTP_OK("UCTP_OK");


//////////////////////////////////////////////////////////////////////////

#define ULOG_CLOUD_PROTOCOL "cloud-protocol"
static LogCategory log(ULOG_CLOUD_PROTOCOL);

void UctpSession::buildQUuid( QUuid* uuid ) const
{
    assert(uuid != NULL);
    
    // construct uuid from MD5 representation of session-id
    
    QByteArray digest = QByteArray::fromHex(uid);
    memcpy((unsigned char*)(uuid), digest.constData(), sizeof(QUuid));
   
    assert(!uuid->isNull());
}

QByteArray Uctp::formRequest(UctpRequestBuilder& requestBuilder )
{
    QByteArray request;
    QXmlStreamWriter stream(&request);
    
    stream.writeStartDocument();
    stream.writeStartElement(UctpElements::REQUEST);
    stream.writeAttribute(UctpAttributes::COMMAND_TYPE, requestBuilder.getCommand());
    
    // Header section
    stream.writeStartElement(UctpElements::HEADER);
    requestBuilder.formHeader(stream);
    stream.writeEndElement();
    
    // Contents section
    stream.writeStartElement(UctpElements::CONTENTS);
    requestBuilder.formContents(stream);
    stream.writeEndDocument();

    return request;
}

bool Uctp::parseReply( QIODevice* reply, const UctpCommand& command, QVariantMap& data)
{
    QXmlInputSource source(reply);

    QList<QString> expectedElements; // = getExpectedElements(command);
    UctpReplyContext context(command, expectedElements);
    UctpReplyHandler* handler = new UctpReplyHandler(context,data);
    xmlReader.setContentHandler(handler);
    xmlReader.setErrorHandler(handler);
    bool ok = xmlReader.parse(source);
    errorString = handler->errorString();
    
    return ok;
}

//////////////////////////////////////////////////////////////////////////

UctpReplyHandler::UctpReplyHandler(const UctpReplyContext& ctx, QVariantMap& data) 
 : replyResultOk(false), context(ctx), replyData(data), envelope(false), header(false), contents(false)
{

}

bool UctpReplyHandler::startElement( const QString &namespaceURI, const QString &localName, const QString &qName, const QXmlAttributes &attributes )
{
    if (qName == UctpElements::RESPONSE) {
        envelope = true;
        UctpCommand cmd = attributes.value(QString(UctpAttributes::COMMAND_TYPE)).toAscii();
        // make sure that response and request have the same type 
        if (cmd != context.getCommand()) {
            errorMessage = QString("Request (%1) and reply (%2) are not the same!")
                .arg(context.getCommand().constData()).arg(cmd.constData());
            return false;
        }
    } else if (qName == UctpElements::HEADER) {
        header = true;     
    }  else if (qName == UctpElements::CONTENTS) {
        contents = true;
    } else if (qName == UctpElements::DATA) {
        dataEntryName = attributes.value(QString(UctpAttributes::NAME));
        if (dataEntryName.isEmpty()) 
        {
            errorMessage = QString("Uctp element 'data' doesn't have NAME attribute");
            return false;
        }
    }
    log.trace(QString("Parsing element %1").arg(qName));
    curText.clear();
    return true;
    
}

bool UctpReplyHandler::endElement(const QString &namespaceURI, const QString &localName, const QString &qName) {
    if (qName == UctpElements::RESULT) {
        if (curText == UctpValues::UCTP_OK) {
            replyResultOk = true;
            log.trace("Uctp request succeed");
        } 
    } else if (qName == UctpElements::DATA) {
        replyData.insert(dataEntryName, curText);
    } else if ( !curText.isEmpty() ){
        replyData.insert(qName, curText);
    } 
    return true;
}

bool UctpReplyHandler::endDocument() {
    bool structureValid  =  header && contents && envelope;
    if (!structureValid) {
        errorMessage = QString(QObject::tr("Invalid response structure"));
        return false;
    }
    if (!validateContext()) {
        errorMessage = QString(QObject::tr("Invalid response: missing elements"));
        return false;
    }

    if (!replyResultOk) {
        QString reason;
        if (replyData.contains(UctpElements::REASON)) {
            reason = replyData.value(UctpElements::REASON).toString();
        } else {
            reason = QString("status is not available");
        }
        errorMessage = QString("Uctp request failed: %1.").arg(reason);
        return false;
    }
    return true;
}



bool UctpReplyHandler::characters(const QString &str) {
    curText += str;
    return true;
}

bool UctpReplyHandler::validateContext() {
    foreach (const QString& elemName, context.getExpectedElements()) {
        if(!replyData.contains(elemName)) {
            return false;
        }
    }
    
    return true;
}

QString UctpReplyHandler::errorString() const {
    if (errorMessage.isEmpty()) {
        return QXmlDefaultHandler::errorString();
    } else {
        return errorMessage;
    }
}


UctpError::~UctpError() throw()
{

}

} // namespace GB2
