/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#ifndef _CLOUD_TRANSPORT_PROTOCOL_H_
#define _CLOUD_TRANSPORT_PROTOCOL_H_

#include <stdexcept>

#include <QtCore/QByteArray>
#include <QtCore/QVariantMap>
#include <QtCore/QIODevice>
#include <QtCore/QUuid>

#include <QtXml/QXmlDefaultHandler>
#include <QtXml/QXmlSimpleReader>
#include <QtXml/QXmlStreamWriter>



namespace GB2 {

// UGene Cloud Transport Protocol (UCTP) definitions

typedef QByteArray UctpCommand;

class UctpElements {
public:
    static const QByteArray REQUEST;
    static const QByteArray RESPONSE;
    static const QByteArray HEADER;
    static const QByteArray CONTENTS;
    static const QByteArray RESULT;
    static const QByteArray REASON;
    static const QByteArray SESSION_ID;
    static const QByteArray USER_NAME;
    static const QByteArray PASSWD;
    static const QByteArray DATA;
    static const QByteArray PROPERTY;
    static const QByteArray TASK_SCHEMA; 
    static const QByteArray TASK_ID;
    static const QByteArray TASK_STATE;
    static const QByteArray TASK_PROGRESS;
    static const QByteArray TASK_ERROR;
};

class UctpAttributes {
public:
     static const QByteArray COMMAND_TYPE;
     static const QByteArray IS_GZIPPED;
     static const QByteArray ENCODING;
     static const QByteArray NAME;
     static const QByteArray SIZE;
};

class UctpValues {
public:
    static const QByteArray BASE64;
    static const QByteArray UCTP_OK;
    static const QByteArray UCTP_TRUE;
    static const QByteArray UCTP_FALSE;
    static const QByteArray UCTP_STATE_FINISHED;
    static const QByteArray UCTP_STATE_ERROR;
};

class UctpCommands {
public:
    static const UctpCommand AUTH;
    static const UctpCommand RUN_TASK;
    static const UctpCommand CANCEL_TASK;
    static const UctpCommand GET_TASK_PROPERTY;
    static const UctpCommand GET_TASK_RESULT;
    static const UctpCommand PING;
};

class UctpReplyContext {
public:
    UctpReplyContext(const UctpCommand& cmd, const QList<QString>& elements )
        : command(cmd), expectedElements(elements) {}
    bool hasElement(const QString& name) { return expectedElements.contains(name); }
    const UctpCommand& getCommand() const {return command;}
    const QList<QString>& getExpectedElements() const {return expectedElements;}
private:
    UctpCommand command;
    QList<QString> expectedElements;
};

class UctpReplyHandler : public QXmlDefaultHandler {
    QString curText;
    QString errorMessage;
    QString dataEntryName;
    bool replyResultOk;
    const UctpReplyContext context;
    QVariantMap& replyData;
    // validation flags present
    bool envelope, header, contents;
public:
    UctpReplyHandler(const UctpReplyContext& ctx, QVariantMap& replyData);
    bool startElement(const QString &namespaceURI, const QString &localName,
        const QString &qName, const QXmlAttributes &attributes);
    bool endElement(const QString &namespaceURI, const QString &localName,
          const QString &qName);
    bool endDocument();
    bool characters(const QString &str);
    QString errorString() const;
private:
    bool validateContext();
};

class UctpRequestBuilder {
    UctpCommand command;
public:
    UctpRequestBuilder(const UctpCommand& cmd) : command(cmd) {}
    const UctpCommand& getCommand(){ return command; }
    virtual void formHeader(QXmlStreamWriter& writer) = 0;
    virtual void formContents(QXmlStreamWriter& writer) = 0;
};

class UctpSession {
public:
    UctpSession(const QByteArray& id) : uid(id) {}
    const QByteArray& getUid() const { return uid; }
    void buildQUuid(QUuid* uuid) const;
private:
    QByteArray uid;
};

class UctpError : public std::runtime_error
{
public:
    UctpError(QString message) : std::runtime_error(message.toAscii().constData()) {}
    const QString getErrorMessage() {return QString( what() ); }
    ~UctpError() throw();
private:
    QString errorMessage;
};

class Uctp {
    QXmlSimpleReader xmlReader;
    QString errorString;
public:
    QByteArray formRequest(UctpRequestBuilder& requestBuilder);
    bool parseReply(QIODevice* reply, const UctpCommand& command, QVariantMap& replyData);
    QString getErrorText() { return errorString; } 
};


} // namespace GB2

#endif // _CLOUD_TRANSPORT_PROTOCOL_H_
