/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "DotPlotSplitter.h"
#include "DotPlotWidget.h"

#include <QtGui/QVBoxLayout>
#include <QtGui/QAction>
#include <QtGui/QMenu>

#include <QtGui/QToolButton>
//#include <QtGui/QToolBar>
#include <util_ov_annotated_dna/ADVSingleSequenceWidget.h>
#include <util_ov_annotated_dna/AnnotatedDNAView.h>

namespace GB2 {

DotPlotSplitter::DotPlotSplitter(AnnotatedDNAView* a)
: ADVSplitWidget(a), locked(false)
{
	QLayout *layout = new QHBoxLayout;
	if (!layout) {
		return;
	}

	layout->setSpacing(0);
	layout->setContentsMargins(0,0,3,0);

	zoomInIcon = QIcon(":core/images/zoom_in.png");
	zoomToIcon = QIcon(":core/images/zoom_sel.png");

	syncLockButton = createToolButton(":core/images/sync_lock.png", tr("Multiple view synchronization lock"), SLOT(sl_toggleSyncLock(bool)));
	aspectRatioButton = createToolButton(":dotplot/images/aspectRatio.png", tr("Keep aspect ratio"), SLOT(sl_toggleAspectRatio(bool)));
	zoomInButton = createToolButton(zoomInIcon, tr("Zoom in/zoom to selection"), SLOT(sl_toggleZoomIn()), false);
	zoomOutButton = createToolButton(":core/images/zoom_out.png", tr("Zoom out"), SLOT(sl_toggleZoomOut()), false);
	resetZoomingButton = createToolButton(":core/images/zoom_whole.png", tr("Reset zooming"), SLOT(sl_toggleZoomReset()), false);

	syncLockButton->setAutoRaise(true);
	syncLockButton->setAutoFillBackground(true);

	aspectRatioButton->setAutoRaise(true);
	zoomInButton->setAutoRaise(true);
	zoomOutButton->setAutoRaise(true);
	resetZoomingButton->setAutoRaise(true);

	buttonToolBar = new HBar(this);
	if (!buttonToolBar) {
		return;
	}

	buttonToolBar->setOrientation(Qt::Vertical);
	buttonToolBar->setFloatable(false);
	buttonToolBar->setMovable(false);
	buttonToolBar->setStyleSheet("background: ");

	buttonToolBar->addWidget(syncLockButton);
//	buttonToolBar->addWidget(aspectRatioButton); // todo: not implemented yet
	buttonToolBar->addWidget(zoomInButton);
	buttonToolBar->addWidget(zoomOutButton);
	buttonToolBar->addWidget(resetZoomingButton);

	splitter = new QSplitter(Qt::Horizontal);
	if (!splitter) {
		return;
	}

	layout->addWidget(buttonToolBar);
	layout->setAlignment(buttonToolBar, Qt::AlignTop);
	layout->addWidget(splitter);
	setAcceptDrops(false);

	setLayout(layout);
}

QToolButton *DotPlotSplitter::createToolButton(const QIcon& ic, const QString& toolTip, const char *slot, bool checkable) {

	QToolButton *toolButton = new QToolButton(this);
	if (toolButton) {
		toolButton->setIcon(ic);
		toolButton->setToolTip(toolTip);
		toolButton->setFixedWidth(20);
		toolButton->setFixedHeight(20);
		toolButton->setCheckable(checkable);
		if (checkable) {
			connect(toolButton, SIGNAL(toggled(bool)), this, slot);
		}
		else {
			connect(toolButton, SIGNAL(clicked()), this, slot);
		}
	}

	return toolButton;
}

QToolButton *DotPlotSplitter::createToolButton(const QString& iconPath, const QString& toolTip, const char *slot, bool checkable) {

	return createToolButton(QIcon(iconPath), toolTip, slot, checkable);
}

DotPlotSplitter::~DotPlotSplitter() {

	delete syncLockButton;
	delete aspectRatioButton;
	delete zoomInButton;
	delete zoomOutButton;
	delete resetZoomingButton;

	delete buttonToolBar;
	delete splitter;
}

void DotPlotSplitter::addView(DotPlotWidget* view) {

	dotPlotList.append(view);
	Q_ASSERT(splitter);
	splitter->addWidget(view);

	connect(view,
		SIGNAL(si_dotPlotChanged(ADVSequenceObjectContext*, ADVSequenceObjectContext*, float, float, QPointF)),
		SLOT(sl_dotPlotChanged(ADVSequenceObjectContext*, ADVSequenceObjectContext*, float, float, QPointF)));

	connect(view,
		SIGNAL(si_dotPlotSelecting()),
		SLOT(sl_dotPlotSelecting()));

	checkLockButtonState();
}

void DotPlotSplitter::removeView(DotPlotWidget* view) {

	dotPlotList.removeAll(view);
	checkLockButtonState();
}

void DotPlotSplitter::checkLockButtonState() {

	QSet< QPair<QString, QString> > dotPlots;

	bool enableLockButton = false;

	int dotPlotCount = dotPlotList.count();
	foreach (DotPlotWidget *dpWidget, dotPlotList) {

		QString nameX = dpWidget->getXSequenceName();
		QString nameY = dpWidget->getYSequenceName();

		if (!nameX.isEmpty() && !nameY.isEmpty()) {

			QPair<QString, QString> p;
			p.first = nameX;
			p.second = nameY;

			if (dotPlots.contains(p)) {
				enableLockButton = true;
			}
			else {
				dotPlots.insert(p);
			}

			dpWidget->setIgnorePanView(dotPlotCount > 1);
		}
	}

	if (syncLockButton) {
		syncLockButton->setEnabled(enableLockButton);
	}
}

bool DotPlotSplitter::isEmpty() const{

	return dotPlotList.isEmpty();
}

// build popup menu for all DotPlotWidgets
// each DotPlotWidget defines if it should add menu
void DotPlotSplitter::buildPopupMenu(QMenu *m) {

	foreach (DotPlotWidget *w, dotPlotList) {
		Q_ASSERT(w);
		w->buildPopupMenu(m);
	}
}

void DotPlotSplitter::sl_toggleSyncLock(bool l) {

	locked = l;
}

void DotPlotSplitter::sl_toggleAspectRatio(bool aspectRatio) {

	foreach (DotPlotWidget *dpWidget, dotPlotList) {
		dpWidget->setKeepAspectRatio(aspectRatio);
	}
}

void DotPlotSplitter::sl_toggleZoomIn() {

	foreach (DotPlotWidget *dpWidget, dotPlotList) {
		dpWidget->zoomIn();
	}
}

void DotPlotSplitter::sl_toggleZoomOut() {

	foreach (DotPlotWidget *dpWidget, dotPlotList) {
		dpWidget->zoomOut();
	}
}

void DotPlotSplitter::sl_toggleZoomReset() {

	foreach (DotPlotWidget *dpWidget, dotPlotList) {
		dpWidget->zoomReset();
	}
}

void DotPlotSplitter::sl_dotPlotChanged(ADVSequenceObjectContext* sequenceX, ADVSequenceObjectContext* sequenceY, float shiftX, float shiftY, QPointF zoom) {

	checkLockButtonState();

	if (locked) {
		foreach (DotPlotWidget* w, dotPlotList) {
			w->setShiftZoom(sequenceX, sequenceY, shiftX, shiftY, zoom);
		}

		update();
	}
}

void DotPlotSplitter::sl_dotPlotSelecting() {

	bool selecting = false;
	foreach (DotPlotWidget* w, dotPlotList) {
		selecting = w->hasSelection();

		if (selecting) {
			break;
		}
	}

	if (selecting) {
		zoomInButton->setIcon(zoomToIcon);
	}
	else {
		zoomInButton->setIcon(zoomInIcon);
	}
}

} // namespace
