/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "ClustalWWorker.h"
#include "TaskLocalStorage.h"
#include "ClustalWSupport.h"

#include <workflow/IntegralBusModel.h>
#include <workflow/WorkflowEnv.h>
#include <workflow/WorkflowRegistry.h>
#include <workflow_support/CoreDataTypes.h>
#include <workflow_library/BioDatatypes.h>
#include <workflow_library/BioActorLibrary.h>
#include <workflow_support/DelegateEditors.h>
#include <workflow_support/CoreLibConstants.h>
#include <core_api/AppContext.h>
#include <core_api/ExternalToolRegistry.h>
#include <core_api/Log.h>

namespace GB2 {
namespace LocalWorkflow {

static LogCategory log(ULOG_CAT_WD);

/****************************
 * ClustalWWorkerFactory
 ****************************/
const QString ClustalWWorkerFactory::ACTOR_ID("clustalw.align");
const QString GAP_OPEN_PENALTY("gapOpenPenalty");
const QString GAP_EXT_PENALTY("gapExtenstionPenalty");
const QString GAP_DIST("gapDist");
const QString END_GAPS("endGaps");
const QString NO_PGAPS("noPGaps");
const QString NO_HGAPS("noHGaps");
const QString EXT_TOOL_PATH("extToolPath");
const QString TMP_DIR_PATH("tmpDirPath");

void ClustalWWorkerFactory::init() {
    QList<PortDescriptor*> p; QList<Attribute*> a;
    Descriptor ind(CoreLibConstants::IN_PORT_ID, ClustalWWorker::tr("Input MSA"), ClustalWWorker::tr("Multiple sequence alignment to be processed."));
    Descriptor oud(CoreLibConstants::OUT_PORT_ID, ClustalWWorker::tr("Multiple sequence alignment"), ClustalWWorker::tr("Result of alignment."));
    p << new PortDescriptor(ind, BioDataTypes::MULTIPLE_ALIGNMENT_TYPE(), true /*input*/);
    p << new PortDescriptor(oud, BioDataTypes::MULTIPLE_ALIGNMENT_TYPE(), false /*input*/, true /*multi*/);
    Descriptor gop(GAP_OPEN_PENALTY, ClustalWWorker::tr("Gap Open Penalty"),
                   ClustalWWorker::tr("Gap Open Penalty"));
    Descriptor gep(GAP_EXT_PENALTY, ClustalWWorker::tr("Gap Extenstion Penalty"),
                   ClustalWWorker::tr("Gap Extenstion Penalty"));
    Descriptor gd(GAP_DIST, ClustalWWorker::tr("Gap Distance"),
                   ClustalWWorker::tr("Gap separation penalty range"));
    Descriptor eg(END_GAPS, ClustalWWorker::tr("End Gaps"),
                   ClustalWWorker::tr("No end gap separation penalty"));
    Descriptor npg(NO_PGAPS, ClustalWWorker::tr("Residue-specific gaps off"),
                   ClustalWWorker::tr("Residue-specific gaps off"));
    Descriptor nhg(NO_HGAPS, ClustalWWorker::tr("Hydrophilic gaps off"),
                   ClustalWWorker::tr("Hydrophilic gaps off"));
    Descriptor etp(EXT_TOOL_PATH, ClustalWWorker::tr("Tool Path"),
                   ClustalWWorker::tr("External tool path"));
    Descriptor tdp(TMP_DIR_PATH, ClustalWWorker::tr("Temporary directory"),
                   ClustalWWorker::tr("Directory for temporary files"));

    a << new Attribute(gop, CoreDataTypes::NUM_TYPE(), false, QVariant(53.90));
    a << new Attribute(gep, CoreDataTypes::NUM_TYPE(), false, QVariant(8.52));
    a << new Attribute(gd, CoreDataTypes::NUM_TYPE(), false, QVariant(4.42));
    a << new Attribute(eg, CoreDataTypes::BOOL_TYPE(), false, QVariant(false));
    a << new Attribute(npg, CoreDataTypes::BOOL_TYPE(), false, QVariant(false));
    a << new Attribute(nhg, CoreDataTypes::BOOL_TYPE(), false, QVariant(false));
    a << new Attribute(etp, CoreDataTypes::STRING_TYPE(), true, QVariant(AppContext::getExternalToolRegistry()->getByName(CLUSTAL_TOOL_NAME)->getPath()));
    a << new Attribute(tdp, CoreDataTypes::STRING_TYPE(), true, QVariant(AppContext::getExternalToolRegistry()->getTemporaryDir()));

    Descriptor desc(ACTOR_ID, ClustalWWorker::tr("ClustalW alignment"),
        ClustalWWorker::tr("ClustalW is a general purpose multiple sequence alignment program for DNA or proteins. "));
    ActorPrototype* proto = new BusActorPrototype(desc, p, a);

    QMap<QString, PropertyDelegate*> delegates;
    {
        QVariantMap m; m["minimum"] = double(.00); m["maximum"] = double(100.00); m["decimals"] = 2;
        delegates[GAP_OPEN_PENALTY] = new DoubleSpinBoxDelegate(m);
    }
    {
        QVariantMap m; m["minimum"] = double(.00); m["maximum"] = double(10.00); m["decimals"] = 2;
        delegates[GAP_EXT_PENALTY] = new DoubleSpinBoxDelegate(m);
    }
    {
        QVariantMap m; m["minimum"] = double(.00); m["maximum"] = double(10.00); m["decimals"] = 2;
        delegates[GAP_DIST] = new DoubleSpinBoxDelegate(m);
    }
    delegates[EXT_TOOL_PATH] = new URLDelegate("", "executable", false);
    delegates[TMP_DIR_PATH] = new URLDelegate("", "TmpDir", false, true);

    proto->setEditor(new DelegateEditor(delegates));
    proto->setPrompter(new ClustalWPrompter());
    proto->setIconPath(":external_tool_support/images/clustalx.png");
    WorkflowEnv::getProtoRegistry()->registerProto(BioActorLibrary::CATEGORY_ALIGNMENT(), proto);

    DomainFactory* localDomain = WorkflowEnv::getDomainRegistry()->getById(LocalDomainFactory::ID);
    localDomain->registerEntry(new ClustalWWorkerFactory());
}

/****************************
* ClustalWPrompter
****************************/
ClustalWPrompter::ClustalWPrompter(Actor* p) : PrompterBase<ClustalWPrompter>(p) {
    connect(AppContext::getExternalToolRegistry()->getByName(CLUSTAL_TOOL_NAME),SIGNAL(si_pathChanged()),SLOT(sl_pathChanged()));
    connect(AppContext::getExternalToolRegistry(),SIGNAL(si_temporyDirChanged()),SLOT(sl_pathChanged()));
}
QString ClustalWPrompter::composeRichDoc() {
    BusPort* input = qobject_cast<BusPort*>(target->getPort(CoreLibConstants::IN_PORT_ID));
    Actor* producer = input->getProducer(CoreLibConstants::IN_PORT_ID);
    QString producerName = producer ? tr(" from %1").arg(producer->getLabel()) : "";
    QString doc = tr("For each MSA<u>%1</u>, build the alignment using <u>\"ClustalW\"</u> and send it to output.")
        .arg(producerName);

    return doc;
}
void ClustalWPrompter::sl_pathChanged(){
    if(target != NULL){
        if(target->getParameter(EXT_TOOL_PATH)->getAttributeValue<QString>() != AppContext::getExternalToolRegistry()->getByName(CLUSTAL_TOOL_NAME)->getPath()){
            target->setParameter(EXT_TOOL_PATH, QVariant(AppContext::getExternalToolRegistry()->getByName(CLUSTAL_TOOL_NAME)->getPath()));
        }
        if(target->getParameter(TMP_DIR_PATH)->getAttributeValue<QString>() != AppContext::getExternalToolRegistry()->getTemporaryDir()){
            target->setParameter(TMP_DIR_PATH, QVariant(AppContext::getExternalToolRegistry()->getTemporaryDir()));
        }
    }
}
/****************************
* ClustalWWorker
****************************/
ClustalWWorker::ClustalWWorker(Actor* a) : BaseWorker(a), input(NULL), output(NULL) {
}

void ClustalWWorker::init() {
    input = ports.value(CoreLibConstants::IN_PORT_ID);
    output = ports.value(CoreLibConstants::OUT_PORT_ID);
}

bool ClustalWWorker::isReady() {
    return (input && input->hasMessage());
}

Task* ClustalWWorker::tick() {
    Message inputMessage = getMessageAndSetupScriptValues(input);
    cfg.gapOpenPenalty=actor->getParameter(GAP_OPEN_PENALTY)->getAttributeValue<float>();
    cfg.gapExtenstionPenalty=actor->getParameter(GAP_EXT_PENALTY)->getAttributeValue<float>();
    cfg.gapDist=actor->getParameter(GAP_DIST)->getAttributeValue<float>();
    cfg.endGaps=actor->getParameter(END_GAPS)->getAttributeValue<bool>();
    cfg.noHGaps=actor->getParameter(NO_HGAPS)->getAttributeValue<bool>();
    cfg.noPGaps=actor->getParameter(NO_PGAPS)->getAttributeValue<bool>();
    QString path=actor->getParameter(EXT_TOOL_PATH)->getAttributeValue<QString>();
    if(path != AppContext::getExternalToolRegistry()->getByName(CLUSTAL_TOOL_NAME)->getPath()){
        AppContext::getExternalToolRegistry()->getByName(CLUSTAL_TOOL_NAME)->setPath(path);
    }
    path=actor->getParameter(TMP_DIR_PATH)->getAttributeValue<QString>();
    if(path != AppContext::getExternalToolRegistry()->getTemporaryDir()){
        AppContext::getExternalToolRegistry()->setTemporaryDir(path);
    }
    MAlignment msa = inputMessage.getData().value<MAlignment>();
    
    if( msa.isEmpty() ) {
        log.error( tr("Empty msa supplied to kalign") );
        return NULL;
    }
    Task* t = new ClustalWSupportTask(new MAlignmentObject(msa), cfg);
    connect(t, SIGNAL(si_stateChanged()), SLOT(sl_taskFinished()));
    return t;
}

void ClustalWWorker::sl_taskFinished() {
    ClustalWSupportTask* t = qobject_cast<ClustalWSupportTask*>(sender());
    if (t->getState() != Task::State_Finished) return;
    QVariant v = qVariantFromValue<MAlignment>(t->resultMA);
    output->put(Message(BioDataTypes::MULTIPLE_ALIGNMENT_TYPE(), v));
    if (input->isEnded()) {
        output->setEnded();
    }
    log.info(tr("Aligned %1 with ClustalW").arg(t->resultMA.getName()));
}

bool ClustalWWorker::isDone() {
    return !input || input->isEnded();
}

void ClustalWWorker::cleanup() {
}

} //namespace LocalWorkflow
} //namespace GB2
