/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "MAFFTWorker.h"
#include "TaskLocalStorage.h"
#include "MAFFTSupport.h"

#include <workflow/IntegralBusModel.h>
#include <workflow/WorkflowEnv.h>
#include <workflow/WorkflowRegistry.h>
#include <workflow_support/CoreDataTypes.h>
#include <workflow_library/BioDatatypes.h>
#include <workflow_library/BioActorLibrary.h>
#include <workflow_support/DelegateEditors.h>
#include <workflow_support/CoreLibConstants.h>
#include <core_api/AppContext.h>
#include <core_api/ExternalToolRegistry.h>
#include <core_api/Log.h>

namespace GB2 {
namespace LocalWorkflow {

static LogCategory log(ULOG_CAT_WD);

/****************************
 * MAFFTWorkerFactory
 ****************************/
const QString MAFFTWorkerFactory::ACTOR_ID("mafft.align");
const QString GAP_OPEN_PENALTY("gapOpenPenalty");
const QString GAP_EXT_PENALTY("gapExtenstionPenalty");
const QString NUM_ITER("maxNumberIterRefinement");
const QString EXT_TOOL_PATH("extToolPath");
const QString TMP_DIR_PATH("tmpDirPath");

void MAFFTWorkerFactory::init() {
    QList<PortDescriptor*> p; QList<Attribute*> a;
    Descriptor ind(CoreLibConstants::IN_PORT_ID, MAFFTWorker::tr("Input MSA"), MAFFTWorker::tr("Multiple sequence alignment to be processed."));
    Descriptor oud(CoreLibConstants::OUT_PORT_ID, MAFFTWorker::tr("Multiple sequence alignment"), MAFFTWorker::tr("Result of alignment."));
    p << new PortDescriptor(ind, BioDataTypes::MULTIPLE_ALIGNMENT_TYPE(), true /*input*/);
    p << new PortDescriptor(oud, BioDataTypes::MULTIPLE_ALIGNMENT_TYPE(), false /*input*/, true /*multi*/);
    Descriptor gop(GAP_OPEN_PENALTY, MAFFTWorker::tr("Gap Open Penalty"),
                   MAFFTWorker::tr("Gap Open Penalty"));
    Descriptor gep(GAP_EXT_PENALTY, MAFFTWorker::tr("Offset"),
                   MAFFTWorker::tr("Works like gap extension penalty"));
    Descriptor tgp(NUM_ITER, MAFFTWorker::tr("Max Iteration"),
                   MAFFTWorker::tr("Maximum number of iterative refinement"));
    Descriptor etp(EXT_TOOL_PATH, MAFFTWorker::tr("Tool Path"),
                   MAFFTWorker::tr("External tool path"));
    Descriptor tdp(TMP_DIR_PATH, MAFFTWorker::tr("Temporary directory"),
                   MAFFTWorker::tr("Directory for temporary files"));

    a << new Attribute(gop, CoreDataTypes::NUM_TYPE(), false, QVariant(1.53));
    a << new Attribute(gep, CoreDataTypes::NUM_TYPE(), false, QVariant(0.00));
    a << new Attribute(tgp, CoreDataTypes::NUM_TYPE(), false, QVariant(0));
    a << new Attribute(etp, CoreDataTypes::STRING_TYPE(), true, QVariant(AppContext::getExternalToolRegistry()->getByName(MAFFT_TOOL_NAME)->getPath()));
    a << new Attribute(tdp, CoreDataTypes::STRING_TYPE(), true, QVariant(AppContext::getExternalToolRegistry()->getTemporaryDir()));

    Descriptor desc(ACTOR_ID, MAFFTWorker::tr("MAFFT alignment"),
        MAFFTWorker::tr("MAFFT is a multiple sequence alignment program for unix-like operating systems. "
                        "<p><dfn>It offers a range of multiple alignment methods, "
                        "L-INS-i (accurate; for alignment of &lt;&#126;200 sequences), "
                        "FFT-NS-2 (fast; for alignment of &lt;&#126;10,000 sequences), etc. </dfn></p>"));
    ActorPrototype* proto = new BusActorPrototype(desc, p, a);

    QMap<QString, PropertyDelegate*> delegates;
    {
        QVariantMap m; m["minimum"] = double(.00); m["maximum"] = double(100.00); m["decimals"] = 2;
        delegates[GAP_OPEN_PENALTY] = new DoubleSpinBoxDelegate(m);
    }
    {
        QVariantMap m; m["minimum"] = double(.00); m["maximum"] = double(10.00); m["decimals"] = 2;
        delegates[GAP_EXT_PENALTY] = new DoubleSpinBoxDelegate(m);
    }
    {
        QVariantMap m; m["minimum"] = int(0); m["maximum"] = int(1000);
        delegates[NUM_ITER] = new DoubleSpinBoxDelegate(m);
    }
    delegates[EXT_TOOL_PATH] = new URLDelegate("", "executable", false);
    delegates[TMP_DIR_PATH] = new URLDelegate("", "TmpDir", false, true);

    proto->setEditor(new DelegateEditor(delegates));
    proto->setPrompter(new MAFFTPrompter());
    proto->setIconPath(":external_tool_support/images/cmdline.png");
    WorkflowEnv::getProtoRegistry()->registerProto(BioActorLibrary::CATEGORY_ALIGNMENT(), proto);

    DomainFactory* localDomain = WorkflowEnv::getDomainRegistry()->getById(LocalDomainFactory::ID);
    localDomain->registerEntry(new MAFFTWorkerFactory());
}

/****************************
* MAFFTPrompter
****************************/
MAFFTPrompter::MAFFTPrompter(Actor* p) : PrompterBase<MAFFTPrompter>(p) {
    connect(AppContext::getExternalToolRegistry()->getByName(MAFFT_TOOL_NAME),SIGNAL(si_pathChanged()),SLOT(sl_pathChanged()));
    connect(AppContext::getExternalToolRegistry(),SIGNAL(si_temporyDirChanged()),SLOT(sl_pathChanged()));
}
QString MAFFTPrompter::composeRichDoc() {
    BusPort* input = qobject_cast<BusPort*>(target->getPort(CoreLibConstants::IN_PORT_ID));
    Actor* producer = input->getProducer(CoreLibConstants::IN_PORT_ID);
    QString producerName = producer ? tr(" from %1").arg(producer->getLabel()) : "";

    QString doc = tr("For each MSA<u>%1</u>, build the alignment using <u>\"MAFFT\"</u> and send it to output.")
        .arg(producerName);

    return doc;
}
void MAFFTPrompter::sl_pathChanged(){
    if(target != NULL){
        if(target->getParameter(EXT_TOOL_PATH)->getAttributeValue<QString>() != AppContext::getExternalToolRegistry()->getByName(MAFFT_TOOL_NAME)->getPath()){
            target->setParameter(EXT_TOOL_PATH, QVariant(AppContext::getExternalToolRegistry()->getByName(MAFFT_TOOL_NAME)->getPath()));
        }
        if(target->getParameter(TMP_DIR_PATH)->getAttributeValue<QString>() != AppContext::getExternalToolRegistry()->getTemporaryDir()){
            target->setParameter(TMP_DIR_PATH, QVariant(AppContext::getExternalToolRegistry()->getTemporaryDir()));
        }
    }
}
/****************************
* MAFFTWorker
****************************/
MAFFTWorker::MAFFTWorker(Actor* a) : BaseWorker(a), input(NULL), output(NULL) {
}

void MAFFTWorker::init() {
    input = ports.value(CoreLibConstants::IN_PORT_ID);
    output = ports.value(CoreLibConstants::OUT_PORT_ID);
}

bool MAFFTWorker::isReady() {
    return (input && input->hasMessage());
}

Task* MAFFTWorker::tick() {
    Message inputMessage = getMessageAndSetupScriptValues(input);
    cfg.gapOpenPenalty=actor->getParameter(GAP_OPEN_PENALTY)->getAttributeValue<float>();
    cfg.gapExtenstionPenalty=actor->getParameter(GAP_EXT_PENALTY)->getAttributeValue<float>();
    cfg.maxNumberIterRefinement=actor->getParameter(NUM_ITER)->getAttributeValue<int>();
    QString path=actor->getParameter(EXT_TOOL_PATH)->getAttributeValue<QString>();
    if(path != AppContext::getExternalToolRegistry()->getByName(MAFFT_TOOL_NAME)->getPath()){
        AppContext::getExternalToolRegistry()->getByName(MAFFT_TOOL_NAME)->setPath(path);
    }
    path=actor->getParameter(TMP_DIR_PATH)->getAttributeValue<QString>();
    if(path != AppContext::getExternalToolRegistry()->getTemporaryDir()){
        AppContext::getExternalToolRegistry()->setTemporaryDir(path);
    }
    MAlignment msa = inputMessage.getData().value<MAlignment>();
    
    if( msa.isEmpty() ) {
        log.error( tr("Empty msa supplied to kalign") );
        return NULL;
    }
    Task* t = new MAFFTSupportTask(new MAlignmentObject(msa), cfg);
    connect(t, SIGNAL(si_stateChanged()), SLOT(sl_taskFinished()));
    return t;
}

void MAFFTWorker::sl_taskFinished() {
    MAFFTSupportTask* t = qobject_cast<MAFFTSupportTask*>(sender());
    if (t->getState() != Task::State_Finished) return;
    QVariant v = qVariantFromValue<MAlignment>(t->resultMA);
    output->put(Message(BioDataTypes::MULTIPLE_ALIGNMENT_TYPE(), v));
    if (input->isEnded()) {
        output->setEnded();
    }
    log.info(tr("Aligned %1 with ClustalW").arg(t->resultMA.getName()));
}

bool MAFFTWorker::isDone() {
    return !input || input->isEnded();
}

void MAFFTWorker::cleanup() {
}

} //namespace LocalWorkflow
} //namespace GB2
