/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "CreateScriptWorker.h"
#include "WorkflowEditorDelegates.h"
#include "library/ScriptWorker.h"

#include <core_api/Log.h>
#include <workflow/WorkflowRegistry.h>

#include <QtCore/QAbstractItemModel>
#include <QtGui/QTableView>
#include <QtGui/QHeaderView>
#include <QtGui/QMessageBox>


namespace GB2 {

#define SETTINGS_ROOT QString("workflow_designer_plugin/")
#define ACTORS_PATH QString("path")

#define WORKFLOW_DOC "GB2WORKFLOW"
#define ACTOR_ELEMENT "Actor"
#define INPUT_PORT_ELEMENT "Input-port"
#define OUTPUT_PORT_ELEMENT "Output-port"
#define ATTRIBUTE_ELEMENT "Attributes"
#define IN_SLOT_ELEMENT "In-Slots"
#define OUT_SLOT_ELEMENT "Out-Slots"
#define SLOT_ID "Slot"
#define ATTR_ELEMENT "Attribute"
#define NAME_ID "Name"
#define TYPE_ID "Type"
#define NAME_ELEMENT "Element-name"
#define DESCR_ELEMENT "Element-description"
#define DESCR_ID "Description"

static LogCategory log(ULOG_CAT_WD);
//Q_DECLARE_METATYPE(DataTypePtr);

class MapForTypesDelegate {
public:
    MapForTypesDelegate() {}

    static QVariantMap getAttrTypes() {
        QVariantMap res;
        DataTypePtr ptr = CoreDataTypes::BOOL_TYPE();
        res[ptr->getDisplayName()] = CoreDataTypes::BOOL_TYPE_ID;

        ptr = CoreDataTypes::NUM_TYPE();
        res[ptr->getDisplayName()] = CoreDataTypes::NUM_TYPE_ID;

        ptr = CoreDataTypes::STRING_TYPE();
        res[ptr->getDisplayName()] = CoreDataTypes::STRING_TYPE_ID;

        return res;
    }

    static QVariantMap getPortTypes() {
        QVariantMap res;
        DataTypePtr ptr = BioDataTypes::DNA_SEQUENCE_TYPE();
        res[ptr->getDisplayName()] = BioDataTypes::DNA_SEQUENCE_TYPE_ID;

        ptr = BioDataTypes::ANNOTATION_TABLE_TYPE();
        res[ptr->getDisplayName()] = BioDataTypes::ANNOTATION_TABLE_TYPE_ID;

        ptr = BioDataTypes::MULTIPLE_ALIGNMENT_TYPE();
        res[ptr->getDisplayName()] = BioDataTypes::MULTIPLE_ALIGNMENT_TYPE_ID;
        return res;
    }

    static QVariantMap portMap;
    static QVariantMap attrMap;
};

QVariantMap MapForTypesDelegate::attrMap = MapForTypesDelegate::getAttrTypes();
QVariantMap MapForTypesDelegate::portMap = MapForTypesDelegate::getPortTypes();


enum DelegateType {
    DelegateForPort,
    DelegateForAttribute
};

class CfgListItem {
public:
    CfgListItem(DelegateType t)  {
        if(t == DelegateForPort) {
            delegate = new ComboBoxDelegate(MapForTypesDelegate::portMap);
            dataTypeId = BioDataTypes::DNA_SEQUENCE_TYPE_ID;
        }
        else if(t == DelegateForAttribute) {
            delegate = new ComboBoxDelegate(MapForTypesDelegate::attrMap);
            dataTypeId = CoreDataTypes::STRING_TYPE_ID;
        }
        else {
            assert(0);
        }
    }
    ~CfgListItem() {delete delegate;}
    PropertyDelegate *getDelegate() const {return delegate;}
    QString getDataType() const {return dataTypeId;}
    void setDataType(const QString& id) {
        dataTypeId = id;
    }

    QString getName() const {return name;}
    void setName(const QString &_name) {name = _name;}

private:
    PropertyDelegate *delegate;
    QString dataTypeId;
    QString name;
    //DataTypePtr dataType;
};


class CfgListModel: public QAbstractListModel {
public:
    CfgListModel(QObject *obj = NULL): QAbstractListModel(obj) {   
        items.append(new CfgListItem(DelegateForPort));
    }

    ~CfgListModel() {
        foreach(CfgListItem *item, items) {
            delete item;
        }
    }

    QStringList getItems() const {
        QStringList result;
        foreach(CfgListItem *item, items) {
            result.append(item->getDataType());
        }
        return result;
    }

    int rowCount(const QModelIndex & parent = QModelIndex()) const {
        return items.count();
    }

    Qt::ItemFlags flags(const QModelIndex &index) const{
        return Qt::ItemIsEditable | Qt::ItemIsEnabled | Qt::ItemIsSelectable;
    }    

    CfgListItem* getItem(const QModelIndex &index) const {
        //CfgListItem *item  = static_cast<CfgListItem*>(index.internalPointer());
        return items.at(index.row());
    }

    QModelIndex parent(const QModelIndex &child) const{
        return QModelIndex();
    }

    QVariant data(const QModelIndex &index, int role /* = Qt::DisplayRole */) const {
        CfgListItem *item = getItem(index);
        PropertyDelegate *dg = item->getDelegate();
        switch(role) {
            case Qt::DisplayRole:
            case Qt::ToolTipRole:
                return dg->getDisplayValue(item->getDataType());
            case DelegateRole:
                return qVariantFromValue<PropertyDelegate*>(dg);
            case Qt::EditRole:
            case ConfigurationEditor::ItemValueRole:
                return item->getDataType();
            default:
                return QVariant();
        }
    }

    bool setData(const QModelIndex &index, const QVariant &value, int role /* = Qt::EditRole */) {
        switch (role)
        {
            case Qt::EditRole:
            case ConfigurationEditor::ItemValueRole:
                CfgListItem * item = getItem(index);
                if(item->getDataType() != value.toString()) {
                    if(!value.toString().isEmpty()) {
                        item->setDataType(value.toString());
                    }
                }
                emit dataChanged(index, index);
        }
        return true;
    }

    bool insertRows ( int row, int count = 0, const QModelIndex & parent = QModelIndex() )  {
        beginInsertRows(parent,items.size(), items.size());
        items.append(new CfgListItem(DelegateForPort));
        endInsertRows();
        return true;
    }

    bool removeRows(int row, int count, const QModelIndex &parent /* = QModelIndex */) {
        if(rowCount() == 1) {
            return false;
        }

        beginRemoveRows(parent, row, row);
        items.removeAt(row);
        endRemoveRows();
        return true;
    }


private:
    QList<CfgListItem*> items;
};

class CfgTableModel: public QAbstractTableModel {
public:
    CfgTableModel(QObject *obj = NULL): QAbstractTableModel(obj) {
        //attrs.append(new CfgListItem());
    }

    int rowCount(const QModelIndex &parent /* = QModelIndex */) const{
        return attrs.size();
    }

    int columnCount(const QModelIndex &parent /* = QModelIndex */) const {
        return 2;
    }

    Qt::ItemFlags flags(const QModelIndex &index) const{
        return Qt::ItemIsEditable | Qt::ItemIsEnabled | Qt::ItemIsSelectable;
    }   

    CfgListItem* getItem(const QModelIndex &index) const {
        //CfgListItem *item  = static_cast<CfgListItem*>(index.internalPointer());
        return attrs.at(index.row());
    }

    QList<CfgListItem*> getItems() const {
        return attrs;
    }

    QVariant data(const QModelIndex &index, int role /* = Qt::DisplayRole */) const {
        CfgListItem *item = getItem(index);
        int col = index.column();
        PropertyDelegate *dg = item->getDelegate();

        switch(role) {
            case Qt::DisplayRole:
                if(col == 0) return item->getName();
                else return dg->getDisplayValue(item->getDataType());
            case DelegateRole:
                if(col == 1)return qVariantFromValue<PropertyDelegate*>(dg);
                else return QVariant();
            case Qt::EditRole:
            case ConfigurationEditor::ItemValueRole:
                if(col == 1) return item->getDataType();
                else return QVariant();
            default:
                return QVariant();
        }
    }

    QVariant headerData ( int section, Qt::Orientation orientation, int role = Qt::DisplayRole ) const {
        if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
            switch(section) {
                case 0: return CreateScriptElementDialog::tr("Name");
                case 1: return CreateScriptElementDialog::tr("type");
                default: return QVariant();
            }
        }
        return QVariant();
    }

    bool setData(const QModelIndex &index, const QVariant &value, int role) {
        int col = index.column();
        CfgListItem * item = getItem(index);
        switch (role)
        {
            case Qt::EditRole:
            case ConfigurationEditor::ItemValueRole:
                if(col == 1) {
                    if(item->getDataType() != value.toString()) {
                        if(!value.toString().isEmpty()) {
                            item->setDataType(value.toString());
                        }
                    }
                }
                else {
                    if(item->getName() != value.toString()) {
                        item->setName(value.toString());
                    }
                }
                emit dataChanged(index, index);
        }
        return true;
    }

    bool insertRows ( int row, int count = 0, const QModelIndex & parent = QModelIndex() )  {
        beginInsertRows(parent, attrs.size(), attrs.size());
        attrs.append(new CfgListItem(DelegateForAttribute));
        endInsertRows();
        return true;
    }

    bool removeRows(int row, int count = 0, const QModelIndex & parent = QModelIndex()) {
        beginRemoveRows(parent, row,row);
        attrs.removeAt(row);
        endRemoveRows();
        return true;
    }

private:
    QList<CfgListItem*>attrs;
};



CreateScriptElementDialog::CreateScriptElementDialog(QWidget *p): QDialog(p) {
    setupUi(this);

    inputList->setModel(new CfgListModel());
    inputList->setItemDelegate(new ProxyDelegate());
    outputList->setModel(new CfgListModel());
    outputList->setItemDelegate(new ProxyDelegate());

    attributeTable->setModel(new CfgTableModel());
    attributeTable->setItemDelegate(new ProxyDelegate());

    connect(addInputButton, SIGNAL(clicked()), SLOT(sl_addInputClicked()));
    connect(addOutputButton, SIGNAL(clicked()), SLOT(sl_addOutputClicked()));
    connect(addAttributeButton, SIGNAL(clicked()), SLOT(sl_addAttribute()));

    connect(deleteInputButton, SIGNAL(clicked()), SLOT(sl_deleteInputClicked()));
    connect(deleteOutputButton, SIGNAL(clicked()), SLOT(sl_deleteOutputClicked()));
    connect(deleteAttributeButton, SIGNAL(clicked()), SLOT(sl_deleteAttributeClicked()));

    connect(okButton, SIGNAL(clicked()), SLOT(sl_okClicked()));
    connect(cancelButton, SIGNAL(clicked()), SLOT(sl_cancelClicked()));

    attributeTable->horizontalHeader()->setStretchLastSection(true);
}

void CreateScriptElementDialog::sl_addInputClicked() {
    inputList->model()->insertRow(0,QModelIndex());
}

void CreateScriptElementDialog::sl_addOutputClicked() {
    outputList->model()->insertRow(0,QModelIndex());
}

void CreateScriptElementDialog::sl_addAttribute() {
    attributeTable->model()->insertRow(0,QModelIndex());
}

void CreateScriptElementDialog::sl_deleteInputClicked() {
    //QModelIndexList indexes = inputList->selectedIndexes();
    QModelIndex index = inputList->currentIndex();
    //foreach (const QModelIndex &index, indexes) {
        inputList->model()->removeRow(index.row());
    //}
}

void CreateScriptElementDialog::sl_deleteOutputClicked() {
    //QModelIndexList indexes = outputList->selectedIndexes();
    QModelIndex index = outputList->currentIndex();
    //foreach (const QModelIndex &index, indexes)  {
        outputList->model()->removeRow(index.row());
    //}
}

void CreateScriptElementDialog::sl_deleteAttributeClicked() {
    //int arow = attributeTable->sele
    //attributeTable->model()->removeRow()
    //QModelIndexList indexes = attributeTable->selectedIndexes(); 
    QModelIndex index = attributeTable->currentIndex();
    //foreach (const QModelIndex &index, indexes) {
        attributeTable->model()->removeRow(index.row());
   // }
}

void CreateScriptElementDialog::sl_cancelClicked() {
    reject();
}

void CreateScriptElementDialog::sl_okClicked() {
    CfgListModel * inputPorts = static_cast<CfgListModel*>(inputList->model());
    QList<QString> typeIds = inputPorts->getItems();
    DataTypeRegistry *dtr = WorkflowEnv::getDataTypeRegistry();
    //input = QList<DataTypePtr>();
    assert(dtr);
    input.clear();
    foreach(const QString & id, typeIds) {
        DataTypePtr ptr = dtr->getById(id);
        if(input.contains(ptr)) {
            QMessageBox::critical(this, tr("error"), tr("Two identical types for input port"));
            log.error(tr("two identical types for input port"));
            return;
        }
        input << ptr;
    }

    CfgListModel * outputPorts = static_cast<CfgListModel*>(outputList->model());
    typeIds = outputPorts->getItems();
    assert(dtr);
    output.clear();
    foreach(const QString & id, typeIds) {
        DataTypePtr ptr = dtr->getById(id);
        if(output.contains(ptr)) {
            QMessageBox::critical(this, tr("error"),tr("Two identical types for output port"));
            log.error(tr("two identical types for output port"));
            return;
        }
        output << ptr;
    }

    CfgTableModel * attrTableModel = static_cast<CfgTableModel*>(attributeTable->model());
    QList<CfgListItem*> attributes = attrTableModel->getItems();
    attrs.clear();
    foreach(CfgListItem *item, attributes) {
        QString itemName = item->getName();
        if(itemName.isEmpty()) {
            QMessageBox::critical(this, tr("error"),tr("Name for some attributes is empty"));
            log.error(tr("Name for some attributes is empty"));
            return;
        }
        foreach(const Attribute *attr, attrs) {
            if(attr->getId() == itemName) {
                QMessageBox::critical(this, tr("error"),tr("Two attributes with name %1").arg(itemName));
                log.error(tr("Two attributes with name %1").arg(itemName));
                return;
            }
        }

        DataTypePtr ptr = dtr->getById(item->getDataType());
        Descriptor desc(itemName, itemName, ptr->getDisplayName());      
        if(ptr == CoreDataTypes::BOOL_TYPE()) {
            attrs << new Attribute(desc, ptr, false, QVariant(false));
        }
        else {
            attrs << new Attribute(desc, ptr);
        }
    }

    name = nameEdit->text();
    if(name.isEmpty()) {
        QMessageBox::critical(this, tr("error"),tr("Name for block is empty"));
        log.error(tr("Name for block is empty"));
        return;
    }

    WProtoRegistry *pr = WorkflowEnv::getProtoRegistry();
    if(pr) {
        if(pr->getProto(LocalWorkflow::ScriptWorkerFactory::ACTOR_ID+name)) {
            QMessageBox::critical(this, tr("error"),tr("Actor with this name already registered"));
            log.error(tr("Actor with this name already registered"));
            return;
        }
    }

    description = descriptionEdit->toPlainText();
    if(description.isEmpty()) {
        QMessageBox::critical(this, tr("error"),tr("Description for block is empty"));
        log.error(tr("Description for block is empty"));
        return;
    }

    saveParams();
    
    accept();
}

void CreateScriptElementDialog::saveParams() {
    QDomDocument doc = saveXml();
    Settings *s = AppContext::getSettings();
    QString defaultPath = QDir::searchPaths( PATH_PREFIX_DATA ).first() + "/workflow_samples/" + "users/";
    QString url = s->getValue(SETTINGS_ROOT + ACTORS_PATH, defaultPath).toString();
    QDir dir(url);
    if(!dir.exists()) {
        dir.mkpath(url);
    }

    IOAdapterFactory *iof = AppContext::getIOAdapterRegistry()->getIOAdapterFactoryById(BaseIOAdapters::LOCAL_FILE);
    IOAdapter *io = iof->createIOAdapter();
    if(io->open( url + name + ".usa", IOAdapterMode_Write )) {
        io->writeBlock(doc.toByteArray());
        io->close();
    }
    else {
        log.error(tr("Can't save user's workflow element"));
    }
}

QDomDocument CreateScriptElementDialog::saveXml() {
    QDomDocument xml(WORKFLOW_DOC);
    QDomElement actor = xml.createElement(ACTOR_ELEMENT);
    xml.appendChild(actor);

    CfgListModel * inputPorts = static_cast<CfgListModel*>(inputList->model());
    QList<QString> typeIds = inputPorts->getItems();
    QDomElement inputPort = xml.createElement(INPUT_PORT_ELEMENT);
    actor.appendChild(inputPort);
    foreach(const QString & str, typeIds) {
        QDomElement slot = xml.createElement(IN_SLOT_ELEMENT);
        slot.setAttribute(SLOT_ID, str);
        inputPort.appendChild(slot);
    }

    CfgListModel * outputPorts = static_cast<CfgListModel*>(outputList->model());
    typeIds = outputPorts->getItems();
    QDomElement outputPort = xml.createElement(OUTPUT_PORT_ELEMENT);
    actor.appendChild(outputPort);
    foreach(const QString & str, typeIds) {
        QDomElement slot = xml.createElement(OUT_SLOT_ELEMENT);
        slot.setAttribute(SLOT_ID, str);
        outputPort.appendChild(slot);
    }
    
    CfgTableModel * attrTableModel = static_cast<CfgTableModel*>(attributeTable->model());
    QList<CfgListItem*> attributes = attrTableModel->getItems();
    QDomElement attribute = xml.createElement(ATTRIBUTE_ELEMENT);
    actor.appendChild(attribute);
    foreach(CfgListItem *item, attributes) {
        QString itemName = item->getName();
        QString itemId = item->getDataType();
        QDomElement attr = xml.createElement(ATTR_ELEMENT);
        attr.setAttribute(NAME_ID, itemName);
        attr.setAttribute(TYPE_ID, itemId);
        attribute.appendChild(attr);
    }

    QDomElement nameEl = xml.createElement(NAME_ELEMENT);
    nameEl.setAttribute(NAME_ID, name);
    actor.appendChild(nameEl);

    QDomElement descriptionEl = xml.createElement(DESCR_ELEMENT);
    descriptionEl.setAttribute(DESCR_ID, description);
    actor.appendChild(descriptionEl);

    return xml;
}

QList<DataTypePtr > CreateScriptElementDialog::getInput() const{
    return input;
}
QList<DataTypePtr > CreateScriptElementDialog::getOutput() const{
    return output;
}
QList<Attribute *> CreateScriptElementDialog::getAttributes() const {
    return attrs;
}
const QString CreateScriptElementDialog::getName() const {
    return name;
}

const QString CreateScriptElementDialog::getDescription() const {
    return description;
}

}//namespace
