/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/


#include <QtXml/QDomDocument>
#include <QtCore/QFile>
#include <QtCore/QTimer>

#include <core_api/L10n.h>
#include <core_api/Log.h>
#include <core_api/Counter.h>

#include <distributed_computing/SerializeUtils.h>
#include <workflow_support/WorkflowIOTasks.h>
#include <workflow_support/SchemaSerializer.h>
#include <workflow_support/CoreLibConstants.h>
#include <workflow_support/CoreDataTypes.h>
#include <time.h>

#include "WorkflowOnTheCloudTask.h"


namespace GB2 {

#define WORKFLOW_ON_CLOUD_LOG_CAT "Workflow-on-cloud task"
static LogCategory log(WORKFLOW_ON_CLOUD_LOG_CAT);

/***************************************
* WorkflowOnTheCloudRunTask
***************************************/

WorkflowOnTheCloudRunTask::WorkflowOnTheCloudRunTask( RemoteMachineSettings * m, const Schema & sc, const QList<Iteration> & its ) 
: Task( tr( "Workflow run task on the cloud" ), TaskFlags_FOSCOE ), machineSettings( m ), 
machine( NULL ), schema( sc ), iterations( its ), taskId(0), eventLoop(NULL) 
{
    GCOUNTER(cvar, tvar, "WorkflowOnTheCloud");
    if( NULL == machineSettings ) {
        setError( tr("Bad remote machine settings"));
        return;
    }
    tpm = Progress_Manual;
}

void WorkflowOnTheCloudRunTask::preprocessSchema()
{
    foreach( Actor * actor, schema.procs ) {
        assert( NULL != actor );
        if( actor->getParameter( CoreLibConstants::URL_IN_ATTR_ID ) != NULL &&
            actor->getParameter( CoreLibConstants::URL_LOCATION_ATTR_ID ) == NULL ) {

                actor->addParameter( CoreLibConstants::URL_LOCATION_ATTR_ID, 
                    new Attribute( CoreLibConstants::URL_LOCATION_ATTR(), CoreDataTypes::BOOL_TYPE(), false, true ) );
        }

        QList<Iteration>::iterator it = iterations.begin();
        ActorId id = actor->getId();
        while( it != iterations.end() ) {
            QList<QString> parameterNames = actor->getParameters().keys();
            foreach( const QString & paramName, parameterNames ) {
                if( !it->cfg[id].contains( paramName ) ) {
                    it->cfg[id][paramName] = actor->getParameter( paramName )->getAttributePureValue();
                }
            }
            ++it;
        }
    }
}

static void dumpSchema(const QString& fileName, const QByteArray& schema) {
    QFile file(fileName);
    file.open(QIODevice::WriteOnly);
    file.write(schema);
    file.close();
}


void WorkflowOnTheCloudRunTask::prepare()
{
    if( hasErrors() || isCanceled() ) {
        return;
    }
    
    log.info(tr("Started remote workflow task."));
    machine = AppContext::getProtocolInfoRegistry()->getProtocolInfo( machineSettings->getProtocolId() )
        ->getRemoteMachineFactory()->createInstance( machineSettings );
    if( NULL == machine ) {
        setError( tr( "Cannot create remote machine from remote machine settings: %1" ).arg( machineSettings->getInfo() ) );
        return;
    }

    preprocessSchema();

    QStringList inputUrls;
    
    foreach( Actor * actor, schema.procs ) {
        assert( NULL != actor );
        ActorId actorId = actor->getId();

        Attribute * urlInAttr = actor->getParameter( CoreLibConstants::URL_IN_ATTR_ID );
        if( NULL != urlInAttr ) {
            QList<Iteration>::iterator it = iterations.begin();
            while( it != iterations.end() ) {
                if( it->cfg[actorId].value( CoreLibConstants::URL_LOCATION_ATTR_ID ).value<bool>() ) { // file located on this computer
                    QString urlpath = it->cfg[actorId].value( CoreLibConstants::URL_IN_ATTR_ID ).value<QString>();
                    // multiple urls are in the same string
                    // TODO: folder contents handling
                    QStringList urls = urlpath.split(';', QString::SkipEmptyParts); 
                    QString newPath;
                    foreach (const GUrl& filePath, urls) {
                        QString path = filePath.fileName();
                        inputUrls.append(filePath.getURLString());
                        newPath.append(";" + path);
                    }
                    // skip first semicolon
                    newPath = newPath.mid(1);
                    it->cfg[actorId][CoreLibConstants::URL_IN_ATTR_ID] = newPath;
                }
                ++it;
            }
        }

        Attribute * urlOutAttr = actor->getParameter( CoreLibConstants::URL_OUT_ATTR_ID );
        if( NULL != urlOutAttr ) {
            assert( NULL == actor->getParameter( CoreLibConstants::URL_LOCATION_ATTR_ID ) );
            QList<Iteration>::iterator it = iterations.begin();
            while( it != iterations.end() ) {
                QVariantMap cfg = it->getParameters( actorId );
                GUrl filePath = cfg.value( CoreLibConstants::URL_OUT_ATTR_ID ).value<QString>();
                QString newPath = filePath.fileName();
                outputUrls.append(filePath.getURLString());
                it->cfg[actorId][CoreLibConstants::URL_OUT_ATTR_ID] = newPath;
                ++it;
            }
        }
    }

    schema.iterations = iterations;
    QByteArray rawData = WorkflowIOUtils::schema2DomDocument(&schema).toByteArray();
    QVariantMap settings;
    settings.insert(CoreLibConstants::WORKFLOW_SCHEMA_ATTR, rawData);
    settings.insert(CoreLibConstants::DATA_IN_ATTR, inputUrls);
    settings.insert(CoreLibConstants::DATA_OUT_ATTR, outputUrls);
    
    log.trace(tr("Schema is preprocessed for sending to cloud"));

#ifdef _DEBUG
    dumpSchema("dump.uws", rawData);
#endif
    
    RemoteTaskError result = machine->runRemoteTask(CoreLibConstants::WORKFLOW_ON_CLOUD_TASK_ID, settings, &taskId);
     if (!result.getOk()) {
         setError(QString("Can not launch remote task: %1").arg(result.getMsg()) );
     }

     

     stateInfo.progress = 0;

}


void WorkflowOnTheCloudRunTask::run()
{
    if (hasErrors() || isCanceled() ) {
        return;
    }
       
    eventLoop = new QEventLoop(this);
    
    QTimer::singleShot( RemoteTask::TIMER_UPDATE_TIME, this, SLOT( sl_remoteTaskTimerUpdate() ) );
    eventLoop->exec(QEventLoop::ExcludeUserInputEvents);
    
    delete eventLoop;
   
}


void WorkflowOnTheCloudRunTask::setRemoteMachineError( const QString& error )
{
    assert(eventLoop != NULL);
    setError(error);
    eventLoop->exit();
}

void WorkflowOnTheCloudRunTask::sl_remoteTaskTimerUpdate() 
{
    assert( eventLoop != NULL );
    if (isCanceled()) {
        RemoteTaskError res = machine->cancelRemoteTask(taskId);
        if (!res.getOk()) {
            setError(tr("Can't cancel remote task: %1").arg( res.getMsg() ));
        }
        
        eventLoop->exit();
        return;
    }  
    
    
    State state = State_Running;
    RemoteTaskError res = machine->getRemoteTaskState(taskId, &state);
    if (!res.getOk()) {
        setRemoteMachineError(res.getMsg());
        return;
    }

    if (state == State_Finished) {
        
        log.info("Workflow task finished on remote host.");
        QVariant resultUrls = outputUrls;
        RemoteTaskError err = machine->getRemoteTaskResult( taskId, &resultUrls );
        if( !err.getOk() ) {
            setRemoteMachineError( tr( "Cannot get remote task result: '%1'" ).arg( err.getMsg() ) );
            return;
        }
        log.info("Retrieved result data from remote host.");
        eventLoop->exit();
        return;

    }

    int progress = 0;
    RemoteTaskError err = machine->getRemoteTaskProgress(taskId, &progress);
    if (!err.getOk()) {
        setRemoteMachineError( tr("Can't get remote task progress: %1").arg(err.getMsg()));
        return;
    }

    stateInfo.progress = progress;

    QTimer::singleShot( RemoteTask::TIMER_UPDATE_TIME, this, SLOT( sl_remoteTaskTimerUpdate() ) );

}


Task::ReportResult WorkflowOnTheCloudRunTask::report()
{
    if (!hasErrors() && !isCanceled()) {
        log.info("Remote task finished successfully");
        log.info("");
    }
    return ReportResult_Finished;
}

WorkflowOnTheCloudRunTask::~WorkflowOnTheCloudRunTask()
{
    delete machine;
}



} // ~GB2


