/*****************************************************************
* Unipro UGENE - Integrated Bioinformatics Suite
* Copyright (C) 2008,2009 Unipro, Russia (http://ugene.unipro.ru)
* All Rights Reserved
* 
*     This source code is distributed under the terms of the
*     GNU General Public License. See the files COPYING and LICENSE
*     for details.
*****************************************************************/

#include "KalignWorker.h"
#include "KalignConstants.h"
#include "TaskLocalStorage.h"
#include "KalignDialogController.h"

#include <workflow/IntegralBusModel.h>
#include <workflow/WorkflowEnv.h>
#include <workflow/WorkflowRegistry.h>
#include <workflow_support/CoreDataTypes.h>
#include <workflow_library/BioDatatypes.h>
#include <workflow_library/BioActorLibrary.h>
#include <workflow_support/DelegateEditors.h>
#include <workflow_support/CoreLibConstants.h>
#include <core_api/Log.h>

/* TRANSLATOR GB2::LocalWorkflow::KalignWorker */

namespace GB2 {
namespace LocalWorkflow {

static LogCategory log(ULOG_CAT_WD);

/****************************
 * KalignWorkerFactory
 ****************************/
const QString KalignWorkerFactory::ACTOR_ID("kalign.align");
const QString GAP_OPEN_PENALTY("gapOpenPenalty");
const QString GAP_EXT_PENALTY("gapExtenstionPenalty");
const QString TERM_GAP_PENALTY("termGapPenalty");
const QString BONUS_SCORE("bonusScore");

void KalignWorkerFactory::init() {
    QList<PortDescriptor*> p; QList<Attribute*> a;
    Descriptor ind(CoreLibConstants::IN_PORT_ID, KalignWorker::tr("Input MSA"), KalignWorker::tr("Multiple sequence alignment to be processed."));
    Descriptor oud(CoreLibConstants::OUT_PORT_ID, KalignWorker::tr("Multiple sequence alignment"), KalignWorker::tr("Result of alignment."));
    p << new PortDescriptor(ind, BioDataTypes::MULTIPLE_ALIGNMENT_TYPE(), true /*input*/);
    p << new PortDescriptor(oud, BioDataTypes::MULTIPLE_ALIGNMENT_TYPE(), false /*input*/, true /*multi*/);
    Descriptor gop(GAP_OPEN_PENALTY, KalignWorker::tr("Gap Open Penalty"),
                   KalignWorker::tr("Gap Open Penalty"));
    Descriptor gep(GAP_EXT_PENALTY, KalignWorker::tr("Gap Extenstion Penalty"),
                   KalignWorker::tr("Gap Extenstion Penalty"));
    Descriptor tgp(TERM_GAP_PENALTY, KalignWorker::tr("Terminal Gap Penalty"),
                   KalignWorker::tr("Terminal Extenstion Penalty"));
//    Descriptor bs(BONUS_SCORE, KalignWorker::tr("Bonus Score"),
//                   KalignWorker::tr("Bonus Score"));

    a << new Attribute(gop, CoreDataTypes::NUM_TYPE(), false, QVariant(53.90));
    a << new Attribute(gep, CoreDataTypes::NUM_TYPE(), false, QVariant(8.52));
    a << new Attribute(tgp, CoreDataTypes::NUM_TYPE(), false, QVariant(4.42));
//    a << new Attribute(bs, CoreDataTypes::NUM_TYPE(), false, QVariant(0.00));

    Descriptor desc(ACTOR_ID, KalignWorker::tr("Kalign alignment"),
        KalignWorker::tr("Kalign is a fast and accurate multiple sequence alignment algorithm designed to align large numbers of protein sequences."
        "<p><dfn>Check <a href=\"http://msa.sbc.su.se\">http://msa.sbc.su.se</a> for the original version.</dfn></p>"));
    ActorPrototype* proto = new BusActorPrototype(desc, p, a);

    QMap<QString, PropertyDelegate*> delegates;
    {
        QVariantMap m; m["minimum"] = double(.00); m["maximum"] = double(100.00); m["decimals"] = 2;
        delegates[GAP_OPEN_PENALTY] = new DoubleSpinBoxDelegate(m);
    }
    {
        QVariantMap m; m["minimum"] = double(.00); m["maximum"] = double(10.00); m["decimals"] = 2;
        delegates[GAP_EXT_PENALTY] = new DoubleSpinBoxDelegate(m);
    }
    {
        QVariantMap m; m["minimum"] = double(.00); m["maximum"] = double(99.99); m["decimals"] = 2;
        delegates[TERM_GAP_PENALTY] = new DoubleSpinBoxDelegate(m);
    }
//    {
//        QVariantMap m; m["minimum"] = double(.00); m["maximum"] = double(10.00); m["decimals"] = 2;
//        delegates[BONUS_SCORE] = new DoubleSpinBoxDelegate(m);
//    }

    proto->setEditor(new DelegateEditor(delegates));
    proto->setPrompter(new KalignPrompter());
    proto->setIconPath(":kalign/images/kalign_16.png");
    WorkflowEnv::getProtoRegistry()->registerProto(BioActorLibrary::CATEGORY_ALIGNMENT(), proto);

    DomainFactory* localDomain = WorkflowEnv::getDomainRegistry()->getById(LocalDomainFactory::ID);
    localDomain->registerEntry(new KalignWorkerFactory());
}

/****************************
* KalignPrompter
****************************/
QString KalignPrompter::composeRichDoc() {
    BusPort* input = qobject_cast<BusPort*>(target->getPort(CoreLibConstants::IN_PORT_ID));
    Actor* producer = input->getProducer(CoreLibConstants::IN_PORT_ID);
    QString producerName = producer ? tr(" from %1").arg(producer->getLabel()) : "";

    QString doc = tr("For each MSA<u>%1</u>, build the alignment using <u>\"Kalign\"</u> and send it to output.")
        .arg(producerName);

    return doc;
}

/****************************
* KalignWorker
****************************/
KalignWorker::KalignWorker(Actor* a) : BaseWorker(a), input(NULL), output(NULL) {
}

void KalignWorker::init() {
    input = ports.value(CoreLibConstants::IN_PORT_ID);
    output = ports.value(CoreLibConstants::OUT_PORT_ID);
}

bool KalignWorker::isReady() {
    return (input && input->hasMessage());
}

Task* KalignWorker::tick() {
    Message inputMessage = getMessageAndSetupScriptValues(input);
    cfg.gapOpenPenalty=actor->getParameter(GAP_OPEN_PENALTY)->getAttributeValue<float>();
    cfg.gapExtenstionPenalty=actor->getParameter(GAP_EXT_PENALTY)->getAttributeValue<float>();
    cfg.termGapPenalty=actor->getParameter(TERM_GAP_PENALTY)->getAttributeValue<float>();

    MAlignment msa = inputMessage.getData().value<MAlignment>();
    
    if( msa.isEmpty() ) {
        log.error( tr("Empty msa supplied to kalign") );
        return NULL;
    }
    Task* t = new KalignTask(msa, cfg);
    connect(t, SIGNAL(si_stateChanged()), SLOT(sl_taskFinished()));
    return t;
}

void KalignWorker::sl_taskFinished() {
    KalignTask* t = qobject_cast<KalignTask*>(sender());
    if (t->getState() != Task::State_Finished) return;
    QVariant v = qVariantFromValue<MAlignment>(t->resultMA);
    output->put(Message(BioDataTypes::MULTIPLE_ALIGNMENT_TYPE(), v));
    if (input->isEnded()) {
        output->setEnded();
    }
    log.info(tr("Aligned %1 with Kalign").arg(t->resultMA.getName()));
}

bool KalignWorker::isDone() {
    return !input || input->isEnded();
}

void KalignWorker::cleanup() {
}

} //namespace LocalWorkflow
} //namespace GB2
